#pragma once

#include <array>
#include <tuple>
#include <type_traits>

namespace Dune
{
  template <class T, int N>
  class FieldVector;

  template <class T, int N, int M>
  class FieldMatrix;
}

namespace AMDiS
{
  // Get the number of elements in a tuple / pair / array / ...
  //@{
  namespace Impl
  {
    template <class Tuple, class = void>
    struct SizeImpl : std::integral_constant<std::size_t, 0> {};

    template <class... Args>
    struct SizeImpl<std::tuple<Args...>>
        : std::integral_constant<std::size_t, sizeof...(Args) > {};

    template <class Arg0, class Arg1>
    struct SizeImpl<std::pair<Arg0, Arg1>>
        : std::integral_constant<std::size_t, 2> {};

    template <class T, std::size_t N>
    struct SizeImpl<std::array<T,N>>
        : std::integral_constant<std::size_t, N> {};

    template <class T, int N>
    struct SizeImpl<Dune::FieldVector<T,N>>
        : std::integral_constant<std::size_t, N> {};

    template <class T, int N, int M>
    struct SizeImpl<Dune::FieldMatrix<T,N,M>>
        : std::integral_constant<std::size_t, N*M> {};

    // Specialization for arithmetic types
    template <class T>
    struct SizeImpl<T, std::enable_if_t<std::is_arithmetic<T>::value> >
        : std::integral_constant<std::size_t, 1> {};

  } // end namespace Impl
  //@}


  template <class T>
  constexpr std::size_t Size = Impl::SizeImpl<std::decay_t<T>>::value;

} // end namespace AMDiS
